# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####


bl_info = {
    'name': "KTools",
    'author': "Kjartan Tysdal",
    'location': '"Shift+Q" and also in EditMode "W-Specials/ KTools"',
    'description': "Adds my personal collection of small handy scripts (mostly modeling tools)",
    'category': "Mesh",
    'blender': (2, 7, 8),
    'version': (0, 2, 9),
    'wiki_url': 'http://www.kjartantysdal.com/scripts',
}

import bpy, bmesh
from bpy.props import StringProperty, IntProperty, FloatProperty, EnumProperty, BoolProperty, \
    BoolVectorProperty, FloatVectorProperty


def check_scale():
    """Check if scale is 0 on any of the axis, if it is then set it to 0.01"""

    y = -1
    for x in bpy.context.object.scale:
        y += 1
        if x == 0.0:
            bpy.context.object.scale[y] = 0.01


# Sybren TODO: double-check this works, could have off-by-one errors.
def layers(layer_nr: int):
    return (layer_nr - 1) * [False] + [True] + (20 - layer_nr) * [False]


class MESH_OT_context_merge(bpy.types.Operator):
    """Context Sensitive Merge"""
    bl_idname = "mesh.context_merge"
    bl_label = "Merge Tool"
    bl_options = {'REGISTER', 'UNDO'}
    
    merge_distance = FloatProperty(
        name="Merge Distance", 
        description="Change the Merge Distance threshold for Remove Doubles.", 
        default=0.0001,
        min=0)

    @classmethod    
    def poll(cls, context):
        return context.active_object.mode == 'EDIT'

    def execute(self, context):

        sel_mode = bpy.context.tool_settings.mesh_select_mode
        
        ob = bpy.context.object
        ob.update_from_editmode() 
        vert_sel = len([v for v in ob.data.vertices if v.select])
        vert_total = len(ob.data.vertices)
        
        if sel_mode[0] and vert_sel <= 2:
            bpy.ops.mesh.merge(type='LAST')
        elif sel_mode[1] and (vert_sel != vert_total):
            bpy.ops.mesh.merge(type='COLLAPSE')
        else:
            bpy.ops.mesh.remove_doubles(threshold=self.merge_distance)

        return {'FINISHED'}


# Adds "Lattice to Selection" to the Addon
class MESH_OT_lattice_to_selection(bpy.types.Operator):
    """Add a lattice deformer to the selection"""
    bl_idname = "object.lattice_to_selection"
    bl_label = "Lattice to Selection"
    bl_options = {'REGISTER', 'UNDO'}

    # Global vs Local
    is_local = BoolProperty(
        name="Local",
        description="Orient the lattice to the active object",
        default=True)
    parent_to = BoolProperty(
        name="Parent to Lattice",
        description="Parents all the objects to the Lattice", default=False)
    move_first = BoolProperty(
        name="First in Modifier Stack",
        description="Moves the lattice modifier to be first in the stack",
        default=False)
    interpolation = bpy.props.EnumProperty(
        items=(('KEY_LINEAR', 'Linear', 'Linear Interpolation'),
               ('KEY_CARDINAL', 'Cardinal', 'Cardinal Interpolation'),
               ('KEY_CATMULL_ROM', 'Catmull Rom', 'Catmull Rom Interpolation'),
               ('KEY_BSPLINE', 'BSpline', 'BSpline Interpolation')),
        name="Interpolation",
        default='KEY_BSPLINE')
    seg_u = IntProperty(name="Lattice U", default=2, soft_min=2)
    seg_v = IntProperty(name="Lattice V", default=2, soft_min=2)
    seg_w = IntProperty(name="Lattice W", default=2, soft_min=2)

    @classmethod
    def poll(cls, context) -> bool:
        mode = context.active_object.mode
        if mode not in {'EDIT', 'OBJECT'}:
            return False
        return True

    def execute(self, context):

        apply_rot = not self.is_local
        parent_to = self.parent_to  # Parents all the objects to the Lattice
        move_first = self.move_first  # moves the lattice modifier to be first in the stack
        interpolation = self.interpolation

        # check if there exists an active object
        if context.scene.objects.active.select:
            active_obj = context.scene.objects.active
        else:
            for ob in context.selected_objects:
                if ob.type == 'MESH':
                    context.scene.objects.active = ob
                    active_obj = ob
                    break
            else:
                self.report({'ERROR'}, 'No active object, and no mesh object selected.')
                return {'CANCELLED'}

        if active_obj.type != 'MESH':
            self.report({'ERROR'}, "Make sure the active object is a Mesh")
            return {'CANCELLED'}

        mode = context.active_object.mode

        # Sybren: try to move duplicated code out of the if/else blocks.
        if mode == 'OBJECT':

            # check if object type is not MESH and then deselect it
            for ob in context.selected_objects:
                if bpy.data.objects[ob.name].type != 'MESH':
                    bpy.data.objects[ob.name].select = False

            org_objs = context.selected_objects

            bpy.ops.object.duplicate()

            # remove any modifiers
            if context.object.modifiers:
                for ob in context.object.modifiers:
                    bpy.ops.object.modifier_remove(modifier=ob.name)

            if len(context.selected_objects) > 1:
                bpy.ops.object.join()

            # create tmp:object and store its location, rotation and dimensions
            bpy.ops.object.transform_apply(location=False, rotation=apply_rot, scale=True)
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY', center='BOUNDS')

            lattice_loc = context.object.location
            lattice_rot = context.object.rotation_euler
            bbox_size = context.object.dimensions
            tmp_obj = context.object.name

            # create the lattice object with the lattice_loc and rot
            bpy.ops.object.add(radius=1, type='LATTICE', view_align=False, enter_editmode=False,
                               location=lattice_loc, rotation=lattice_rot)

            lattice_obj = context.object

            # set dimensions / bounding box size
            context.object.scale = bbox_size

            bpy.ops.object.select_all(action='DESELECT')

            # select and delete the tmp_object
            bpy.data.objects[tmp_obj].select = True
            bpy.ops.object.delete(use_global=False)

            # select all the original objects and assign the lattice deformer
            for i in org_objs:
                if bpy.data.objects[i.name].type == 'MESH':
                    context.scene.objects.active = bpy.data.objects[i.name]
                    bpy.data.objects[i.name].select = True

                    bpy.ops.object.modifier_add(type='LATTICE')
                    lattice_name = context.object.modifiers[
                        len(context.object.modifiers) - 1].name
                    context.object.modifiers[lattice_name].object = lattice_obj
                    if move_first:
                        for ob in context.object.modifiers:
                            bpy.ops.object.modifier_move_up(modifier=lattice_name)
                else:
                    bpy.data.objects[i.name].select = True

            if parent_to:

                bpy.data.objects[lattice_obj.name].select = True
                context.scene.objects.active = bpy.data.objects[lattice_obj.name]

                bpy.ops.object.parent_set(type='OBJECT', keep_transform=True)
            else:

                bpy.ops.object.select_all(action='DESELECT')
                bpy.data.objects[lattice_obj.name].select = True
                context.scene.objects.active = bpy.data.objects[lattice_obj.name]

        elif mode == 'EDIT':

            org_objs = context.selected_objects

            # Add vertex group and store its name in a variable
            bpy.ops.object.vertex_group_assign_new()
            v_id = len(context.object.vertex_groups) - 1
            context.object.vertex_groups[v_id].name = 'tmp_lattice_to_selection'
            v_group = context.object.vertex_groups[v_id].name

            bpy.ops.mesh.duplicate()
            bpy.ops.mesh.separate(type='SELECTED')

            bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

            for ob in context.selected_objects:
                if ob not in org_objs:
                    tmp_obj = ob.name

            bpy.ops.object.select_all(action='DESELECT')

            context.scene.objects.active = bpy.data.objects[tmp_obj]
            bpy.data.objects[tmp_obj].select = True

            if context.object.modifiers:
                for ob in context.object.modifiers:
                    bpy.ops.object.modifier_remove(modifier=ob.name)

            bpy.ops.object.transform_apply(location=False, rotation=apply_rot, scale=True)
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY', center='BOUNDS')

            lattice_loc = context.object.location
            lattice_rot = context.object.rotation_euler
            bbox_size = context.object.dimensions
            tmp_obj = context.object.name

            bpy.ops.object.add(radius=1, type='LATTICE', view_align=False, enter_editmode=False,
                               location=lattice_loc, rotation=lattice_rot)

            lattice_obj = context.object

            context.object.scale = bbox_size

            bpy.ops.object.select_all(action='DESELECT')

            bpy.data.objects[tmp_obj].select = True

            bpy.ops.object.delete(use_global=False)

            context.scene.objects.active = bpy.data.objects[active_obj]
            bpy.data.objects[active_obj].select = True

            bpy.ops.object.modifier_add(type='LATTICE')
            lattice_name = context.object.modifiers[len(context.object.modifiers) - 1].name
            context.object.modifiers[lattice_name].object = lattice_obj
            context.object.modifiers[lattice_name].vertex_group = v_group

            if move_first:
                for ob in context.object.modifiers:
                    bpy.ops.object.modifier_move_up(modifier=lattice_name)

            bpy.ops.object.select_all(action='DESELECT')

            bpy.data.objects[lattice_obj.name].select = True
            context.scene.objects.active = bpy.data.objects[lattice_obj.name]
        else:
            raise ValueError('Mode is unexpected %r' % mode)

        context.object.data.interpolation_type_u = interpolation
        context.object.data.interpolation_type_v = interpolation
        context.object.data.interpolation_type_w = interpolation

        context.object.data.points_u = self.seg_u
        context.object.data.points_v = self.seg_v
        context.object.data.points_w = self.seg_w

        check_scale()

        return {'FINISHED'}

    def invoke(self, context, event):
        wm = context.window_manager
        return wm.invoke_props_dialog(self)


# Adds Calculate Normals and Smooth to the Addon
class MESH_OT_calc_normals(bpy.types.Operator):
    """Calculates and smooths normals."""
    bl_idname = "mesh.calc_normals"
    bl_label = "Calculate Normals"
    bl_options = {'REGISTER', 'UNDO'}

    invert = BoolProperty(
        name="Invert Normals", 
        description="Inverts the normals.", 
        default=False)
    
    @classmethod    
    def poll(cls, context):
        mode = bpy.context.active_object.mode
        return mode == 'OBJECT' or mode == 'EDIT'
    
    def execute(self, context):

        invert = self.invert
        mode = context.active_object.mode

        if mode == 'OBJECT':

            sel = context.selected_objects
            active = context.scene.objects.active.name

            bpy.ops.object.shade_smooth()

            for ob in sel:
                ob = ob.name
                context.scene.objects.active = bpy.data.objects[ob]
                bpy.ops.object.editmode_toggle()
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.mesh.normals_make_consistent(inside=invert)
                bpy.ops.object.editmode_toggle()

            context.scene.objects.active = bpy.data.objects[active]

        elif mode == 'EDIT':
            bpy.ops.mesh.normals_make_consistent(inside=invert)

        return {'FINISHED'}


# Adds SnapToAxis to the Addon
class MESH_OT_snaptoaxis(bpy.types.Operator):
    """Snaps selected vertices to zero on the selected axis."""
    bl_idname = "mesh.snaptoaxis"
    bl_label = "Snap to Axis"
    bl_options = {'REGISTER', 'UNDO'}


    snap_x = BoolProperty(
        name="Snap to X",
        description="Snaps to zero in X. Also sets the axis for the mirror modifier if that button is turned on",
        default=True)
    snap_y = BoolProperty(
        name="Snap to Y",
        description="Snaps to zero in Y. Also sets the axis for the mirror modifier if that button is turned on",
        default=False)
    snap_z = BoolProperty(
        name="Snap to Z",
        description="Snaps to zero in Z. Also sets the axis for the mirror modifier if that button is turned on",
        default=False)

    mirror_add = BoolProperty(
        name="Add Mirror Modifier", 
        description="Adds a mirror modifer",
        default=False)

    mirror_x = BoolProperty(
        name="Mirror on X", 
        description="Sets the modifier to mirror on X",
        default=True)
    mirror_y = BoolProperty(
        name="Mirror on Y", 
        description="Sets the modifier to mirror on Y",
        default=False)
    mirror_z = BoolProperty(
        name="Mirror on Z", 
        description="Sets the modifier to mirror on Z",
        default=False)
    clipping = BoolProperty(
        name="Enable Clipping",
        description="Prevents vertices from going through the mirror during transform",
        default=True)

    def draw(self, context):
        layout = self.layout
        col = layout.column()

        col_move = col.column(align=True)
        row = col_move.row(align=True)
        if self.snap_x:
            row.prop(self, "snap_x", text="X", icon='CHECKBOX_HLT')
        else:
            row.prop(self, "snap_x", text="X", icon='CHECKBOX_DEHLT')
        if self.snap_y:
            row.prop(self, "snap_y", text="Y", icon='CHECKBOX_HLT')
        else:
            row.prop(self, "snap_y", text="Y", icon='CHECKBOX_DEHLT')
        if self.snap_z:
            row.prop(self, "snap_z", text="Z", icon='CHECKBOX_HLT')
        else:
            row.prop(self, "snap_z", text="Z", icon='CHECKBOX_DEHLT')

        col.separator()

        col_move = col.column(align=True)
        col_move.prop(self, "mirror_add", icon='MODIFIER')
        row = col_move.row(align=True)

        row = col_move.row(align=True)
        row.active = self.mirror_add
        if self.mirror_x:
            row.prop(self, "mirror_x", text="X", icon='CHECKBOX_HLT')
        else:
            row.prop(self, "mirror_x", text="X", icon='CHECKBOX_DEHLT')
        if self.mirror_y:
            row.prop(self, "mirror_y", text="Y", icon='CHECKBOX_HLT')
        else:
            row.prop(self, "mirror_y", text="Y", icon='CHECKBOX_DEHLT')
        if self.mirror_z:
            row.prop(self, "mirror_z", text="Z", icon='CHECKBOX_HLT')
        else:
            row.prop(self, "mirror_z", text="Z", icon='CHECKBOX_DEHLT')

        col = col.column()
        col.active = self.mirror_add
        col.prop(self, "clipping")
    
    @classmethod
    def poll(cls, context):
        mode = bpy.context.active_object.mode
        return mode == 'OBJECT' or mode == 'EDIT'

    def execute(self, context):

        mode = context.active_object.mode
        mirror_find = context.object.modifiers.find('Mirror')
        run = True

        if mode == 'EDIT':
            loc = context.object.location

            me = context.object.data
            bm = bmesh.from_edit_mesh(me)

            for v in bm.verts:
                if v.select:
                    if self.snap_x:
                        v.co.x = 0
                    if self.snap_y:
                        v.co.y = 0
                    if self.snap_z:
                        v.co.z = 0

            bmesh.update_edit_mesh(me, True, False)

        if self.mirror_add:

            if mirror_find <= -1:
                bpy.ops.object.modifier_add(type='MIRROR')

                context.object.modifiers['Mirror'].show_viewport = True

                run = False

            context.object.modifiers["Mirror"].use_clip = self.clipping
            context.object.modifiers["Mirror"].use_x = self.mirror_x
            context.object.modifiers["Mirror"].use_y = self.mirror_y
            context.object.modifiers["Mirror"].use_z = self.mirror_z

        return {'FINISHED'}



# Adds Autotubes to the Addon
class MESH_OT_autotubes(bpy.types.Operator):
    """Creates a spline tube based on selected edges"""
    bl_idname = "mesh.autotubes"
    bl_label = "Auto Tubes"
    bl_options = {'REGISTER', 'UNDO'}

    bevel = FloatProperty(name="Tube Width", description="Change width of the tube.", default=0.1,
                          min=0)
    res = IntProperty(name="Tube Resolution", description="Change resolution of the tube.",
                      default=2, min=0, max=20)
    
    @classmethod
    def poll(cls, context):
        mode = bpy.context.active_object.mode
        return mode == 'EDIT'
    
    def execute(self, context):

        mode = context.active_object.mode
        type = context.active_object.type
        bevel = self.bevel
        res = self.res

        if mode == 'EDIT' and type == 'MESH':
            bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
            bpy.ops.object.duplicate()

            bpy.ops.object.mode_set(mode='EDIT', toggle=False)
            bpy.ops.mesh.select_all(action='INVERT')
            bpy.ops.mesh.delete(type='EDGE')

            bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
            bpy.ops.object.subdivision_set(level=0)
            bpy.ops.object.convert(target='CURVE')
            context.object.data.fill_mode = 'FULL'
            context.object.data.bevel_depth = 0.1
            context.object.data.splines[0].use_smooth = True
            context.object.data.bevel_resolution = 2
            bpy.ops.object.shade_smooth()

            bpy.ops.object.mode_set(mode='EDIT', toggle=False)
            bpy.ops.curve.spline_type_set(type='BEZIER')

            context.object.data.bevel_depth = bevel
            context.object.data.bevel_resolution = res



        elif type == 'CURVE':

            context.object.data.bevel_depth = bevel
            context.object.data.bevel_resolution = res

        elif mode != 'EDIT' and type == 'MESH':
            self.report({'ERROR'}, "This one only works in Edit mode")
            return {'CANCELLED'}

        return {'FINISHED'}



class MESH_OT_context_cut(bpy.types.Operator):
    """Context sensitive Cut tool"""
    bl_idname = "mesh.context_cut"
    bl_label = "Cut Tool"
    bl_options = {'REGISTER', 'UNDO'}

    cuts = IntProperty(
        name="Number of Cuts", 
        description="Change the number of cuts.", 
        default=1,
        min=1, 
        soft_max=10)
    loopcut = BoolProperty(
        name="Insert LoopCut",
        description="Makes a loop cut based on the selected edges",
        default=False)
    smoothness = FloatProperty(
        name="Smoothness", 
        description="Change the smoothness.", 
        default=0,
        min=0, 
        soft_max=1)
    quad_corners = bpy.props.EnumProperty(
        items=(
            ('INNERVERT', 'Inner Vert', 'How to subdivide quad corners'),
            ('PATH', 'Path', 'How to subdivide quad corners'),
            ('STRAIGHT_CUT', 'Straight Cut', 'How to subdivide quad corners'),
            ('FAN', 'Fan', 'How to subdivide quad corners')),
        name="Quad Corner Type", default='STRAIGHT_CUT')
    
    @classmethod
    def poll(cls, context):
        mode = bpy.context.active_object.mode
        return mode == 'EDIT'

    def execute(self, context):

        quad_corners = self.quad_corners
        cuts = self.cuts
        loopcut = self.loopcut
        smoothness = self.smoothness
        mode = context.active_object.mode

        if mode == 'EDIT':
            sel_mode = context.tool_settings.mesh_select_mode[:]

            # Checks and stores if any Vert, Edge or Face is selected.
            bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

            me = context.object.data
            bm = bmesh.new()  # create an empty BMesh
            bm.from_mesh(me)  # fill it in from a Mesh
            sel = []
            edge_sel = []
            vert_sel = []

            for v in bm.faces:
                if v.select:
                    sel.append(v.index)
            for v in bm.edges:
                if v.select:
                    edge_sel.append(v.index)
            for v in bm.verts:
                if v.select:
                    vert_sel.append(v.index)

            bm.to_mesh(me)
            bpy.ops.object.mode_set(mode='EDIT', toggle=False)


            if sel_mode[2] and len(sel) > 1:

                vgrp = context.object.vertex_groups.active_index

                # Store the Hidden Polygons
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.object.vertex_group_assign_new()
                tmp_hidden = context.object.vertex_groups.active_index
                bpy.ops.mesh.select_all(action='DESELECT')

                # Select faces to be cut
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                mesh = context.active_object.data.polygons

                for f in sel:
                    mesh[f].select = True

                bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                bpy.ops.mesh.hide(unselected=True)
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.mesh.region_to_loop()

                # Store Boundry Edges
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

                me = context.object.data
                bm = bmesh.new()
                bm.from_mesh(me)
                boundry_edge = []

                for v in bm.edges:
                    if v.select:
                        boundry_edge.append(v.index)

                bm.to_mesh(me)

                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

                # Store Cut Edges
                bpy.ops.mesh.select_all(action='INVERT')
                bpy.ops.mesh.loop_multi_select(ring=True)

                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

                me = context.object.data
                bm = bmesh.new()
                bm.from_mesh(me)
                cut_edges = []

                for v in bm.edges:
                    if v.select:
                        cut_edges.append(v.index)

                bm.to_mesh(me)

                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

                # Store Intersection edges
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='FACE')
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='EDGE')

                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

                me = context.object.data
                bm = bmesh.new()
                bm.from_mesh(me)
                int_edges = []

                for v in bm.edges:
                    if v.select:
                        int_edges.append(v.index)

                bm.to_mesh(me)

                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

                # Modify Lists
                for x in int_edges:
                    if x in boundry_edge:
                        cut_edges.remove(x)

                bpy.ops.mesh.select_all(action='DESELECT')

                # Select the new edges to cut
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                mesh = context.active_object.data.edges

                for f in cut_edges:
                    mesh[f].select = True

                # Perform cut and select the cut line.
                bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                bpy.ops.mesh.subdivide(number_cuts=cuts, smoothness=smoothness,
                                       quadcorner=quad_corners)
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.mesh.region_to_loop()
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='VERT')
                bpy.ops.mesh.select_all(action='INVERT')
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='EDGE')
                bpy.ops.mesh.loop_multi_select(ring=False)

                # Store cut line.
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

                me = context.object.data
                bm = bmesh.new()
                bm.from_mesh(me)
                cut_line = []

                for v in bm.edges:
                    if v.select:
                        cut_line.append(v.index)

                bm.to_mesh(me)

                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

                bpy.ops.mesh.reveal()
                bpy.ops.mesh.select_all(action='DESELECT')

                context.object.vertex_groups.active_index = tmp_hidden
                bpy.ops.object.vertex_group_select()
                bpy.ops.mesh.hide(unselected=True)

                bpy.ops.mesh.select_all(action='DESELECT')

                # Select Cutline
                if cuts <= 1:
                    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                    mesh = context.active_object.data.edges

                    for f in cut_line:
                        mesh[f].select = True

                    bpy.ops.object.mode_set(mode='EDIT', toggle=False)

                bpy.ops.object.vertex_group_remove(all=False)
                bpy.ops.mesh.select_mode(use_extend=True, use_expand=False, type='FACE')

            elif sel_mode[0] and len(vert_sel) >= 2:
                bpy.ops.mesh.vert_connect_path()

            elif sel_mode[1] and not loopcut and edge_sel:
                bpy.ops.mesh.subdivide(number_cuts=cuts, smoothness=smoothness,
                                       quadcorner=quad_corners)

            elif sel_mode[1] and loopcut and edge_sel:
                bpy.ops.mesh.loop_multi_select(ring=True)
                bpy.ops.mesh.subdivide(number_cuts=cuts, smoothness=smoothness,
                                       quadcorner=quad_corners)
            else:
                bpy.ops.mesh.select_all(action='DESELECT')
                bpy.ops.mesh.knife_tool("INVOKE_DEFAULT")

        else:
            self.report({'ERROR'}, "This one only works in Edit mode")

        return {'FINISHED'}


# Adds custom_autosmooth to the Addon

class MESH_OT_custom_autosmooth(bpy.types.Operator):
    """Set AutoSmooth angle"""
    bl_idname = "object.custom_autosmooth"
    bl_label = "Custom Autosmooth"
    bl_options = {'REGISTER', 'UNDO'}

    angle = FloatProperty(name="AutoSmooth Angle", description="Set AutoSmooth angle", default=30.0,
                          min=0.0, max=180.0)
    
    @classmethod
    def poll(cls, context):
        mode = bpy.context.active_object.mode
        return mode == 'OBJECT' or mode == 'EDIT'

    def execute(self, context):

        mode = context.active_object.mode

        if mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

        ob = context.selected_objects
        angle = self.angle
        angle = angle * (3.14159265359 / 180)

        bpy.ops.object.shade_smooth()

        for x in ob:
            x = x.name
            bpy.data.objects[x].data.use_auto_smooth = True
            bpy.data.objects[x].data.auto_smooth_angle = angle

        bpy.ops.object.mode_set(mode=mode, toggle=False)

        return {'FINISHED'}


# Adds shrinkwrapSmooth to the Addon

class MESH_OT_shrinkwrap_smooth(bpy.types.Operator):
    """Smooths the selected vertices while trying to keep the original shape with a shrinkwrap modifier. """
    bl_idname = "mesh.shrinkwrap_smooth"
    bl_label = "Shrinkwrap Smooth"
    bl_options = {'REGISTER'}

    pin = BoolProperty(
        name="Pin Selection Border",
        description="Pins the outer edge of the selection.", 
        default=True)
    subsurf = IntProperty(
        name="Subsurf Levels", 
        description="More reliable, but slower results",
        default=0, min=0, soft_max=4)
    
    @classmethod
    def poll(cls, context):
        mode = bpy.context.active_object.mode
        return mode == 'EDIT'

    def execute(self, context):
        
        bpy.ops.ed.undo_push() # undo fix
        
        iterate = 6
        pin = self.pin
        data = context.object.data.name

        # Set up for vertex weight
        context.scene.tool_settings.vertex_group_weight = 1
        v_grps = len(context.object.vertex_groups.items())

        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
        org_ob = context.object.name

        # Create intermediate object
        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
        bpy.ops.mesh.primitive_plane_add(radius=1, view_align=False, enter_editmode=False)
        context.object.data = bpy.data.meshes[data]
        tmp_ob = context.object.name

        bpy.ops.object.duplicate(linked=False)
        shrink_ob = context.object.name

        bpy.ops.object.select_all(action='DESELECT')
        bpy.ops.object.select_pattern(pattern=tmp_ob)
        context.scene.objects.active = bpy.data.objects[tmp_ob]

        bpy.ops.object.mode_set(mode='EDIT', toggle=False)
        bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='VERT')

        if v_grps >= 1:
            for x in range(v_grps):
                bpy.ops.object.vertex_group_add()

        if pin:
            bpy.ops.object.vertex_group_assign_new()
            org_id = context.object.vertex_groups.active_index

            bpy.ops.object.vertex_group_assign_new()
            sel = context.object.vertex_groups.active.name
            sel_id = context.object.vertex_groups.active_index

            bpy.ops.mesh.region_to_loop()
            bpy.ops.object.vertex_group_remove_from(use_all_groups=False, use_all_verts=False)

            bpy.ops.mesh.select_all(action='SELECT')
            bpy.ops.mesh.region_to_loop()
            bpy.ops.object.vertex_group_remove_from(use_all_groups=False, use_all_verts=False)

            bpy.ops.mesh.select_all(action='DESELECT')
            bpy.ops.object.vertex_group_select(sel_id)


        else:
            bpy.ops.object.vertex_group_assign_new()
            sel = context.object.vertex_groups.active.name

        for x in range(iterate):
            bpy.ops.object.modifier_add(type='SHRINKWRAP')
            mod_id = (len(context.object.modifiers) - 1)
            shrink_name = context.object.modifiers[mod_id].name

            context.object.modifiers[shrink_name].target = bpy.data.objects[shrink_ob]
            context.object.modifiers[shrink_name].vertex_group = sel

            context.object.modifiers[shrink_name].wrap_method = 'PROJECT'
            context.object.modifiers[shrink_name].use_negative_direction = True
            context.object.modifiers[shrink_name].subsurf_levels = self.subsurf

            bpy.ops.mesh.vertices_smooth(factor=1, repeat=1)

            bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
            bpy.ops.object.convert(target='MESH')
            bpy.ops.object.mode_set(mode='EDIT', toggle=False)

        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

        bpy.ops.object.vertex_group_remove(all=False)
        bpy.ops.object.modifier_remove(modifier=shrink_name)

        bpy.ops.object.select_all(action='DESELECT')
        bpy.ops.object.select_pattern(pattern=shrink_ob)
        context.scene.objects.active = bpy.data.objects[shrink_ob]

        # Delete all geo inside Shrink_Object
        bpy.ops.object.mode_set(mode='EDIT', toggle=False)
        bpy.ops.mesh.select_all(action='SELECT')
        bpy.ops.mesh.delete(type='VERT')
        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

        bpy.ops.object.delete(use_global=True)

        bpy.ops.object.select_pattern(pattern=tmp_ob)
        context.scene.objects.active = bpy.data.objects[tmp_ob]

        bpy.ops.object.mode_set(mode='EDIT', toggle=False)

        if pin:
            bpy.ops.mesh.select_all(action='DESELECT')
            bpy.ops.object.vertex_group_select(org_id)

        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

        bpy.ops.object.delete(use_global=False)

        bpy.ops.object.select_pattern(pattern=org_ob)
        context.scene.objects.active = bpy.data.objects[org_ob]

        bpy.ops.object.mode_set(mode='EDIT', toggle=False)

        # Fix for Blender remembering the previous selection
        bpy.ops.object.vertex_group_assign_new()
        bpy.ops.object.vertex_group_remove(all=False)

        return {'FINISHED'}


# Adds buildCorner to the Addon

class MESH_OT_build_corner(bpy.types.Operator):
    """Builds corner topology. Good for converting ngons"""
    bl_idname = "mesh.build_corner"
    bl_label = "Build Corner"
    bl_options = {'REGISTER', 'UNDO'}

    offset = IntProperty()

    @classmethod
    def poll(cls, context):
        mode = bpy.context.active_object.mode
        return mode == 'EDIT'

    def modal(self, context, event):

        if event.type == 'MOUSEMOVE':

            delta = self.offset - event.mouse_x

            if delta >= 0:
                offset = 1
            else:
                offset = 0

            bpy.ops.mesh.edge_face_add()

            bpy.ops.mesh.poke()
            bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='VERT')
            bpy.ops.object.vertex_group_assign_new()
            sel_id = context.object.vertex_groups.active_index

            bpy.ops.mesh.region_to_loop()
            bpy.ops.object.vertex_group_remove_from()

            bpy.ops.mesh.select_nth(nth=2, skip=1, offset=offset)

            bpy.ops.object.vertex_group_select(sel_id)

            bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='EDGE')
            bpy.ops.mesh.dissolve_mode(use_verts=False)

            bpy.ops.mesh.select_all(action='DESELECT')
            bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='VERT')
            bpy.ops.object.vertex_group_select()
            bpy.ops.mesh.select_more()

            bpy.ops.object.vertex_group_remove(all=False)
            bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='FACE')


        elif event.type == 'LEFTMOUSE':
            return {'FINISHED'}

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            bpy.ops.ed.undo()
            return {'CANCELLED'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        if context.object:

            # Check selection

            bpy.ops.mesh.edge_face_add()
            bpy.ops.mesh.region_to_loop()

            bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

            me = context.object.data
            bm = bmesh.new()  # create an empty BMesh
            bm.from_mesh(me)  # fill it in from a Mesh

            face_sel = []
            edge_sel = []

            for v in bm.faces:
                if v.select:
                    face_sel.append(v.index)
            for v in bm.edges:
                if v.select:
                    edge_sel.append(v.index)

            bm.to_mesh(me)
            bpy.ops.object.mode_set(mode='EDIT', toggle=False)
            bpy.ops.mesh.loop_to_region()

            #------------------------------------

            edge_sel = len(edge_sel)

            if edge_sel == 4:
                return {'FINISHED'}

            elif edge_sel % 2 == 0:
                self.offset = event.mouse_x
                context.window_manager.modal_handler_add(self)
                return {'RUNNING_MODAL'}


            else:
                bpy.ops.mesh.poke()
                bpy.ops.mesh.quads_convert_to_tris(quad_method='BEAUTY', ngon_method='BEAUTY')
                bpy.ops.mesh.tris_convert_to_quads(face_threshold=3.14159, shape_threshold=3.14159)
                return {'FINISHED'}

        else:
            self.report({'WARNING'}, "No active object, could not finish")
            return {'CANCELLED'}


class MESH_OT_draw_poly(bpy.types.Operator):
    """Draw a polygon"""
    bl_idname = "mesh.draw_poly"
    bl_label = "Draw Poly"

    cursor_co = FloatVectorProperty()
    vert_count = 0
    manip = BoolProperty()
    vgrp = IntProperty()
    sel_mode = BoolVectorProperty()
    cursor_depth = BoolProperty()
    snap = BoolProperty()
    
    @classmethod
    def poll(cls, context):
        mode = bpy.context.active_object.mode
        return mode == 'OBJECT' or mode == 'EDIT'
    
    def modal(self, context, event):

        # set header gui
        context.area.tag_redraw()
        context.area.header_text_set(
            "LMB = Create New Point, SHIFT = Flip Normals, RMB / ENTER = Accept NGon, MMB = Accept QuadFill")

        mesh = context.active_object.data

        if event.type == 'LEFTMOUSE':
            if event.value == 'PRESS':

                bpy.ops.view3d.cursor3d('INVOKE_DEFAULT')

                obj = context.active_object
                vert_co = context.scene.cursor_location
                world = obj.matrix_world.inverted_safe()

                bpy.ops.mesh.select_all(action='DESELECT')
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

                me = context.object.data
                bm = bmesh.new()
                bm.from_mesh(me)

                # Add new vert
                new_vert = bm.verts.new(vert_co)
                new_vert.co = world * new_vert.co
                new_vert_id = new_vert.index
                self.vert_count += 1

                if self.vert_count >= 2:
                    bm.verts.ensure_lookup_table()
                    set_of_verts = set(bm.verts[i] for i in range(-2, 0))
                    bm.edges.new(set_of_verts)

                # Get index of first and last vertex
                first_index = len(bm.verts) - self.vert_count
                second_index = first_index + 1
                third_index = first_index + 2
                second_to_last_index = len(bm.verts) - 2

                bm.to_mesh(me)
                bm.free()

                mesh.vertices[new_vert_id].select = True
                bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                context.scene.cursor_location = self.cursor_co

                if self.vert_count >= 4:
                    bpy.ops.object.vertex_group_assign()

                if self.vert_count == 3:
                    # remove second vertex from group
                    bpy.ops.mesh.select_all(action='DESELECT')
                    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                    mesh.vertices[first_index].select = True
                    mesh.vertices[third_index].select = True
                    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                    bpy.ops.object.vertex_group_assign()
                    bpy.ops.mesh.select_more()

                if self.vert_count == 2:
                    bpy.ops.mesh.select_more()

                if self.vert_count >= 4:
                    # make core poly
                    bpy.ops.mesh.select_all(action='DESELECT')
                    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                    mesh.vertices[first_index].select = True
                    mesh.vertices[second_index].select = True
                    mesh.vertices[third_index].select = True
                    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                    bpy.ops.mesh.edge_face_add()

                if self.vert_count == 4:
                    bpy.ops.mesh.select_all(action='DESELECT')
                    bpy.ops.object.vertex_group_select(self.vgrp)
                    bpy.ops.mesh.edge_face_add()

                    # Remove remaining core edge
                    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                    mesh.vertices[second_index].select = True
                    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                    bpy.ops.mesh.edge_face_add()

                if self.vert_count >= 5:
                    bpy.ops.mesh.select_all(action='DESELECT')

                    # Remove Last Edge
                    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                    mesh.vertices[first_index].select = True
                    mesh.vertices[second_to_last_index].select = True
                    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                    bpy.ops.mesh.delete(type='EDGE')

                    # Fill in rest of face
                    bpy.ops.object.vertex_group_select(self.vgrp)
                    bpy.ops.mesh.edge_face_add()

                    # Remove remaining core edge
                    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                    mesh.vertices[second_index].select = True
                    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                    bpy.ops.mesh.edge_face_add()
                    bpy.ops.mesh.flip_normals()

        elif event.type == 'MIDDLEMOUSE':

            # reset header gui
            context.area.tag_redraw()
            context.area.header_text_set()

            # Convert to Quads
            bpy.ops.mesh.quads_convert_to_tris(quad_method='BEAUTY', ngon_method='BEAUTY')
            bpy.ops.mesh.tris_convert_to_quads()
            bpy.ops.mesh.tris_convert_to_quads(face_threshold=3.14159, shape_threshold=3.14159)

            # restore selection mode and manipulator
            context.tool_settings.mesh_select_mode = self.sel_mode
            context.space_data.show_manipulator = self.manip
            context.user_preferences.view.use_mouse_depth_cursor = self.cursor_depth
            context.scene.tool_settings.use_snap = self.snap

            # Remove and make sure vertex group data is gone
            bpy.ops.object.vertex_group_remove_from(use_all_verts=True)
            bpy.ops.object.vertex_group_remove()
            bpy.ops.object.vertex_group_assign_new()
            context.object.vertex_groups.active.name = "drawPoly_temp"
            bpy.ops.object.vertex_group_remove()

            return {'CANCELLED'}

        elif event.type in {'RIGHTMOUSE', 'ESC', 'SPACE'}:

            # reset header gui
            context.area.tag_redraw()
            context.area.header_text_set()

            # restore selection mode and manipulator
            context.tool_settings.mesh_select_mode = self.sel_mode
            context.space_data.show_manipulator = self.manip
            context.user_preferences.view.use_mouse_depth_cursor = self.cursor_depth
            context.scene.tool_settings.use_snap = self.snap

            # Remove and make sure vertex group data is gone
            bpy.ops.object.vertex_group_remove_from(use_all_verts=True)
            bpy.ops.object.vertex_group_remove()
            bpy.ops.object.vertex_group_assign_new()
            context.object.vertex_groups.active.name = "drawPoly_temp"
            bpy.ops.object.vertex_group_remove()

            return {'CANCELLED'}

        elif event.type == 'LEFT_SHIFT' or event.type == 'RIGHT_SHIFT':
            bpy.ops.mesh.flip_normals()

            return {'PASS_THROUGH'}

        elif event.type == 'LEFT_CTRL' or event.type == 'RIGHT_CTRL':
            if context.user_preferences.view.use_mouse_depth_cursor:
                context.user_preferences.view.use_mouse_depth_cursor = False
                context.scene.tool_settings.use_snap = False
            else:
                context.user_preferences.view.use_mouse_depth_cursor = True
                context.scene.tool_settings.use_snap = True
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):

        sel_ob = len(context.selected_objects)

        if sel_ob >= 1:
            sel_type = context.object.type

            if sel_type == 'MESH':
                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

            else:
                self.report({'WARNING'}, "Active object is not a mesh.")
                return {'CANCELLED'}


        elif sel_ob == 0:
            bpy.ops.mesh.primitive_plane_add()
            context.selected_objects[0].name = "polyDraw"
            context.selected_objects[0].data.name = "polyDraw"
            bpy.ops.object.mode_set(mode='EDIT', toggle=False)
            bpy.ops.mesh.select_all(action='SELECT')
            bpy.ops.mesh.delete(type='VERT')

        # Store selection mode, snap and manipulator settings
        self.sel_mode = context.tool_settings.mesh_select_mode[:]
        context.tool_settings.mesh_select_mode = True, False, False
        self.manip = context.space_data.show_manipulator
        context.space_data.show_manipulator = False
        self.cursor_depth = context.user_preferences.view.use_mouse_depth_cursor
        context.user_preferences.view.use_mouse_depth_cursor = False
        self.snap = context.scene.tool_settings.use_snap
        context.scene.tool_settings.use_snap = False

        bpy.ops.object.mode_set(mode='EDIT', toggle=False)

        bpy.ops.mesh.select_all(action='DESELECT')
        bpy.ops.object.vertex_group_assign_new()
        self.vgrp = context.object.vertex_groups.active_index
        context.object.vertex_groups.active.name = "drawPoly_temp"

        self.cursor_co = context.scene.cursor_location

        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}


class MESH_OT_toggle_silhouette(bpy.types.Operator):
    """Turns everything black so that you can evaluate the overall shape. Useful when designing"""
    bl_idname = "object.toggle_silhouette"
    bl_label = "Toggle Silhouette"

    diff_col = FloatVectorProperty(default=(0.226, 0.179, 0.141))
    disp_mode = StringProperty(default='SOLID')
    matcap = BoolProperty(default=False)
    only_render = BoolProperty(default=False)

    def execute(self, context):

        light_check = context.user_preferences.system.solid_lights[0].use

        if light_check:
            # Set Lights to Off
            context.user_preferences.system.solid_lights[0].use = False
            context.user_preferences.system.solid_lights[1].use = False

            # Store variables
            self.diff_col = context.user_preferences.system.solid_lights[2].diffuse_color
            self.disp_mode = context.space_data.viewport_shade
            self.matcap = context.space_data.use_matcap
            self.only_render = context.space_data.show_only_render

            context.user_preferences.system.solid_lights[2].diffuse_color = 0, 0, 0
            context.space_data.viewport_shade = 'SOLID'
            context.space_data.use_matcap = False
            context.space_data.show_only_render = True

        else:
            context.user_preferences.system.solid_lights[0].use = True
            context.user_preferences.system.solid_lights[1].use = True
            context.user_preferences.system.solid_lights[2].diffuse_color = self.diff_col
            context.space_data.viewport_shade = self.disp_mode
            context.space_data.use_matcap = self.matcap
            context.space_data.show_only_render = self.only_render

        return {'FINISHED'}


# Adds growLoop to the Addon

class MESH_OT_grow_loop(bpy.types.Operator):
    """Grows the selected edges in both directions """
    bl_idname = "mesh.grow_loop"
    bl_label = "Grow Loop"
    bl_options = {'REGISTER', 'UNDO'}

    grow = IntProperty(
        name="Grow Selection", 
        description="How much to grow selection", 
        default=1,
        min=1, 
        soft_max=10)

    @classmethod
    def poll(cls, context):
        mode = bpy.context.active_object.mode
        return mode == 'EDIT'

    def execute(self, context):

        grow = self.grow
        sel_mode = context.tool_settings.mesh_select_mode

        for x in range(grow):
            if sel_mode[2]:

                edge_sel = []
                border = []
                interior = []
                face_org = []
                face_loop = []
                face_grow = []
                face_sel = []
                mesh_edges = context.active_object.data.edges
                mesh_faces = context.active_object.data.polygons

                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='FACE')

                me = context.object.data
                bm = bmesh.from_edit_mesh(me)

                for e in bm.edges:
                    if e.select:
                        edge_sel.append(e.index)

                for f in bm.faces:
                    if f.select:
                        face_org.append(f.index)

                bpy.ops.mesh.region_to_loop()

                for e in bm.edges:
                    if e.select:
                        border.append(e.index)

                for e in edge_sel:
                    if e not in border:
                        interior.append(e)

                bmesh.update_edit_mesh(me, True, False)

                bpy.ops.mesh.select_all(action='DESELECT')

                # Select the interior edges
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

                for e in interior:
                    mesh_edges[e].select = True

                bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                bpy.ops.mesh.loop_multi_select(ring=True)
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=True, type='FACE')

                me = context.object.data
                bm = bmesh.from_edit_mesh(me)

                for f in bm.faces:
                    if f.select:
                        face_loop.append(f.index)

                bmesh.update_edit_mesh(me, True, False)

                bpy.ops.mesh.select_all(action='DESELECT')

                # Select original faces
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                for x in face_org:
                    mesh_faces[x].select = True
                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

                bpy.ops.mesh.select_more(use_face_step=False)

                me = context.object.data
                bm = bmesh.from_edit_mesh(me)

                for f in bm.faces:
                    if f.select:
                        face_grow.append(f.index)

                for f in face_grow:
                    if f in face_loop:
                        face_sel.append(f)

                for f in face_org:
                    face_sel.append(f)

                bmesh.update_edit_mesh(me, True, False)

                bpy.ops.mesh.select_all(action='DESELECT')

                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

                for f in face_sel:
                    mesh_faces[f].select = True

                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

            else:
                mesh = context.active_object.data.edges

                me = context.object.data
                bm = bmesh.from_edit_mesh(me)
                org_sel = []
                grow_sel = []
                loop_sel = []
                sel = []

                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='EDGE')

                for e in bm.edges:
                    if e.select:
                        org_sel.append(e.index)

                bpy.ops.mesh.select_more(use_face_step=False)

                for e in bm.edges:
                    if e.select:
                        grow_sel.append(e.index)

                bpy.ops.mesh.select_all(action='DESELECT')

                bmesh.update_edit_mesh(me, True, False)

                # Select the original edges
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                for e in org_sel:
                    mesh[e].select = True
                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

                me = context.object.data
                bm = bmesh.from_edit_mesh(me)
                bpy.ops.mesh.loop_multi_select(ring=False)

                for e in bm.edges:
                    if e.select:
                        loop_sel.append(e.index)

                bmesh.update_edit_mesh(me, True, False)

                bpy.ops.mesh.select_all(action='DESELECT')

                for x in loop_sel:
                    if x in grow_sel:
                        sel.append(x)

                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                for e in sel:
                    mesh[e].select = True
                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

            context.tool_settings.mesh_select_mode = sel_mode

        return {'FINISHED'}





# Adds shrinkLoop to the Addon

class MESH_OT_shrink_loop(bpy.types.Operator):
    """Shrink the selected loop"""
    bl_idname = "mesh.shrink_loop"
    bl_label = "Shrink Loop"
    bl_options = {'REGISTER', 'UNDO'}

    shrink = IntProperty(
        name="Shrink Selection", 
        description="How much to shrink selection",
        default=1, 
        min=1, 
        soft_max=15)

    @classmethod
    def poll(cls, context):
        mode = bpy.context.active_object.mode
        return mode == 'EDIT'

    def execute(self, context):

        sel_mode = context.tool_settings.mesh_select_mode
        shrink = self.shrink

        for x in range(shrink):
            if sel_mode[2]:
                me = context.object.data
                bm = bmesh.from_edit_mesh(me)
                mesh = context.active_object.data

                sel = []
                edge_dic = {}
                vert_list = []
                end_verts = []
                org_faces = []
                cur_faces = []
                new_faces = []

                # Store edges and verts
                for e in bm.edges:
                    if e.select:
                        sel.append(e.index)

                        # Populate vert_list
                        vert_list.append(e.verts[0].index)
                        vert_list.append(e.verts[1].index)

                        # Store dictionary
                        edge_dic[e.index] = [e.verts[0].index, e.verts[1].index]

                # Store original faces
                for f in bm.faces:
                    if f.select:
                        org_faces.append(f.index)

                # Store end verts
                for v in vert_list:
                    if vert_list.count(v) == 2:
                        end_verts.append(v)

                # Check verts in dictionary
                for key, value in edge_dic.items():
                    if value[0] in end_verts:
                        sel.remove(key)
                        continue
                    if value[1] in end_verts:
                        sel.remove(key)

                bmesh.update_edit_mesh(me, True, False)

                # Select the resulting edges
                bpy.ops.mesh.select_all(action='DESELECT')
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

                for e in sel:
                    mesh.edges[e].select = True
                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='VERT')
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='FACE')

                bm = bmesh.from_edit_mesh(me)

                # Store current faces
                for f in bm.faces:
                    if f.select:
                        cur_faces.append(f.index)

                # Compare current and original faces
                for x in org_faces:
                    if x in cur_faces:
                        new_faces.append(x)

                bmesh.update_edit_mesh(me, True, False)

                # Select the resulting faces
                bpy.ops.mesh.select_all(action='DESELECT')
                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

                for e in new_faces:
                    mesh.polygons[e].select = True
                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

            else:
                me = context.object.data
                bm = bmesh.from_edit_mesh(me)

                sel = []
                edge_dic = {}
                vert_list = []
                end_verts = []

                # Store edges and verts in dictionary
                for e in bm.edges:
                    if e.select:
                        sel.append(e.index)

                        # Populate vert_list
                        vert_list.append(e.verts[0].index)
                        vert_list.append(e.verts[1].index)

                        # Store dictionary
                        edge_dic[e.index] = [e.verts[0].index, e.verts[1].index]

                # Store end verts
                for v in vert_list:
                    if vert_list.count(v) == 1:
                        end_verts.append(v)

                # Check verts in dictionary
                for key, value in edge_dic.items():
                    if value[0] in end_verts:
                        sel.remove(key)
                        continue
                    if value[1] in end_verts:
                        sel.remove(key)

                bmesh.update_edit_mesh(me, True, False)

                # Select the resulting edges
                bpy.ops.mesh.select_all(action='DESELECT')

                bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                mesh = context.active_object.data.edges
                for e in sel:
                    mesh[e].select = True
                bpy.ops.object.mode_set(mode='EDIT', toggle=False)

        return {'FINISHED'}


class MESH_OT_paint_select(bpy.types.Operator):
    """Click and drag to select"""
    bl_idname = "view3d.select_paint"
    bl_label = "Paint Select"
    bl_options = {'REGISTER', 'UNDO'}

    deselect = BoolProperty(
        default=False, 
        description='Deselect objects, polys, edges or verts')
    toggle = BoolProperty(
        default=False,
        description='Toggles the selection. NOTE: this option can be slow on heavy meshes')
    sel_before = IntProperty(
        description='Do Not Touch', 
        options={'HIDDEN'})
    sel_after = IntProperty(
        description='Do Not Touch', 
        options={'HIDDEN'})
    
    @classmethod
    def poll(cls, context):
        mode = bpy.context.active_object.mode
        return mode == 'OBJECT' or mode == 'EDIT'
    
    def modal(self, context, event):

        refresh = event.mouse_x

        if self.deselect == False:
            bpy.ops.view3d.select('INVOKE_DEFAULT', extend=True, deselect=False)
        else:
            bpy.ops.view3d.select('INVOKE_DEFAULT', extend=False, deselect=True, toggle=True)

        if event.value == 'RELEASE':
            return {'FINISHED'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):

        if self.toggle:
            sel_ob = len(context.selected_objects)

            if sel_ob >= 1:
                mode = context.object.mode

                if mode == 'EDIT':
                    sel_mode = context.tool_settings.mesh_select_mode[:]

                    # Get Selection before
                    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                    ob = context.object.data
                    # check verts
                    if sel_mode[0]:
                        for v in ob.vertices:
                            if v.select:
                                self.sel_before += 1
                    # check edges
                    elif sel_mode[1]:
                        for e in ob.edges:
                            if e.select:
                                self.sel_before += 1
                    # check polys
                    else:
                        for p in ob.polygons:
                            if p.select:
                                self.sel_before += 1

                    # Toggle Selection
                    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                    bpy.ops.view3d.select('INVOKE_DEFAULT', extend=False, toggle=True)

                    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                    ob = context.object.data
                    # check verts after
                    if sel_mode[0]:
                        for v in ob.vertices:
                            if v.select:
                                self.sel_after += 1

                    # check edges after
                    elif sel_mode[1]:
                        for e in ob.edges:
                            if e.select:
                                self.sel_after += 1
                    # check polys after
                    else:
                        for p in ob.polygons:
                            if p.select:
                                self.sel_after += 1

                    if self.sel_after > self.sel_before:
                        self.deselect = False
                    elif self.sel_after == self.sel_before:
                        bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                        bpy.ops.mesh.select_all(action='DESELECT')
                        return {'FINISHED'}
                    else:
                        self.deselect = True

                    bpy.ops.object.mode_set(mode='EDIT', toggle=False)

                elif mode == 'OBJECT':
                    bpy.ops.view3d.select('INVOKE_DEFAULT', extend=False, toggle=True)

                    sel_ob_after = len(context.selected_objects)

                    if sel_ob_after < sel_ob:
                        self.deselect = True

        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}



# Draws the Custom Menu in Object Mode
class MESH_OT_ktools_menu(bpy.types.Menu):
    bl_label = "KTools - Object Mode"
    bl_idname = "OBJECT_MT_ktools_menu"

    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_DEFAULT'
        layout.operator("mesh.draw_poly")
        layout.operator("object.toggle_silhouette")
        layout.operator("mesh.quickbool")
        layout.operator("mesh.calc_normals")
        layout.operator("object.custom_autosmooth")
        layout.operator("object.basic_rename")
        layout.operator("object.lattice_to_selection")


# Draws the Custom Menu in Edit Mode
class VIEW3D_MT_edit_mesh_ktools_menu_edit(bpy.types.Menu):
    bl_label = "KTools - Edit Mode"
    bl_idname = "VIEW3D_MT_edit_mesh_ktools_menu_edit"

    def draw(self, context):
        layout = self.layout

        layout.operator("mesh.context_cut")
        layout.operator("mesh.context_merge")
        layout.operator("mesh.snaptoaxis")
        layout.operator("mesh.autotubes")
        layout.operator("mesh.shrinkwrap_smooth")

        layout.operator_context = 'INVOKE_DEFAULT'
        layout.operator("mesh.build_corner")
        layout.operator("object.lattice_to_selection")

        layout.separator()

        layout.operator("mesh.grow_loop")
        layout.operator("mesh.shrink_loop")

        layout.separator()

        layout.operator("mesh.draw_poly")
        layout.operator("object.toggle_silhouette")
        layout.operator("object.custom_autosmooth")
        layout.operator("mesh.calc_normals")


# Calls the KTools Object Menu
class MESH_OT_ktools(bpy.types.Operator): 
    """Calls the KTools Menu"""
    bl_idname = "object.ktools"
    bl_label = "KTools Object Menu"
    
    @classmethod    
    def poll(cls, context):
        return context.active_object.mode == 'OBJECT'

    def execute(self, context):
        bpy.ops.wm.call_menu(name=MESH_OT_ktools_menu.bl_idname)
        return {'FINISHED'}


# Calls the KTools Edit Menu
class MESH_OT_ktools_mesh(bpy.types.Operator): 
    """Calls the KTools Edit Menu"""
    bl_idname = "mesh.ktools_mesh"
    bl_label = "KTools Mesh Menu"

    @classmethod    
    def poll(cls, context):
        return context.active_object.mode == 'EDIT'

    def execute(self, context):
        bpy.ops.wm.call_menu(name=VIEW3D_MT_edit_mesh_ktools_menu_edit.bl_idname)
        return {'FINISHED'}


# draw function for integration in menus
def menu_func(self, context):
    self.layout.separator()
    self.layout.menu("VIEW3D_MT_edit_mesh_ktools_menu_edit", text="KTools")


# Register and Unregister all the operators
def register():
    bpy.utils.register_class(MESH_OT_context_merge)
    bpy.utils.register_class(MESH_OT_lattice_to_selection)
    bpy.utils.register_class(MESH_OT_calc_normals)
    bpy.utils.register_class(MESH_OT_snaptoaxis)
    bpy.utils.register_class(MESH_OT_autotubes)
    bpy.utils.register_class(MESH_OT_context_cut)
    bpy.utils.register_class(MESH_OT_custom_autosmooth)
    bpy.utils.register_class(MESH_OT_shrinkwrap_smooth)
    bpy.utils.register_class(MESH_OT_build_corner)
    bpy.utils.register_class(MESH_OT_draw_poly)
    bpy.utils.register_class(MESH_OT_toggle_silhouette)
    bpy.utils.register_class(MESH_OT_grow_loop)
    bpy.utils.register_class(MESH_OT_shrink_loop)
    bpy.utils.register_class(MESH_OT_paint_select)
    bpy.utils.register_class(MESH_OT_ktools_menu)
    bpy.utils.register_class(VIEW3D_MT_edit_mesh_ktools_menu_edit)
    bpy.utils.register_class(MESH_OT_ktools)
    bpy.utils.register_class(MESH_OT_ktools_mesh)
    bpy.types.VIEW3D_MT_edit_mesh_specials.append(menu_func)


    kc = bpy.context.window_manager.keyconfigs.addon
    if kc:
        # Add KTools object menu to inputs
        km = kc.keymaps.new(name="Object Mode", space_type="EMPTY")
        kmi = km.keymap_items.new('object.ktools', 'Q', 'PRESS', shift=True)
        # Add KTools edit menu to inputs
        km = kc.keymaps.new(name="Mesh", space_type="EMPTY")
        kmi = km.keymap_items.new('mesh.ktools_mesh', 'Q', 'PRESS', shift=True)
        # Add paint select to CTRL+SHIFT+ALT+LeftMouse
        km = kc.keymaps.new(name="3D View", space_type="VIEW_3D")
        kmi = km.keymap_items.new('view3d.select_paint', 'ACTIONMOUSE', 'PRESS', shift=True,
                                  ctrl=True, alt=True)


def unregister():
    bpy.utils.unregister_class(MESH_OT_context_merge)
    bpy.utils.unregister_class(MESH_OT_lattice_to_selection)
    bpy.utils.unregister_class(MESH_OT_calc_normals)
    bpy.utils.unregister_class(MESH_OT_snaptoaxis)
    bpy.utils.unregister_class(MESH_OT_autotubes)
    bpy.utils.unregister_class(MESH_OT_context_cut)
    bpy.utils.unregister_class(MESH_OT_custom_autosmooth)
    bpy.utils.unregister_class(MESH_OT_shrinkwrap_smooth)
    bpy.utils.unregister_class(MESH_OT_build_corner)
    bpy.utils.unregister_class(MESH_OT_draw_poly)
    bpy.utils.unregister_class(MESH_OT_toggle_silhouette)
    bpy.utils.unregister_class(MESH_OT_grow_loop)
    bpy.utils.unregister_class(MESH_OT_shrink_loop)
    bpy.utils.unregister_class(MESH_OT_paint_select)
    bpy.utils.unregister_class(MESH_OT_ktools_menu)
    bpy.utils.unregister_class(VIEW3D_MT_edit_mesh_ktools_menu_edit)
    bpy.utils.unregister_class(MESH_OT_ktools)
    bpy.utils.unregister_class(MESH_OT_ktools_mesh)
    bpy.types.VIEW3D_MT_edit_mesh_specials.remove(menu_func)

    kc = bpy.context.window_manager.keyconfigs.addon
    if kc:
        km = kc.keymaps["Object Mode"]
        for kmi in km.keymap_items:
            if kmi.idname == 'object.ktools':
                km.keymap_items.remove(kmi)
                break
        km = kc.keymaps["Mesh"]
        for kmi in km.keymap_items:
            if kmi.idname == 'mesh.ktools_mesh':
                km.keymap_items.remove(kmi)
                break
        km = kc.keymaps["3D View"]
        for kmi in km.keymap_items:
            if kmi.idname == 'view3d.select_paint':
                km.keymap_items.remove(kmi)
                break